#include <windows.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include "telldus-core.h"
#include "Ibtmexcw.h"

#define TARGET_TEMP (37.5)
#define DIFF_TEMP (0.0)
#define WARN_TEMP (0.5)
#define ALARM_TEMP (1.0)

static int verbose = 0;

long session_handle;
unsigned char state_buffer[15360];
char status[5000];

static double temperature, temp_target, temp_warn_high, temp_warn_low, temp_crit_high, temp_crit_low;

static short read_scratchpad(void)
{
	short i, res;
	state_buffer[0] = 0xBE;	/* read scratchpad */
	for (i = 1; i < 10; i++) state_buffer[i] = 0xFF;
	res = TMBlockStream(session_handle, state_buffer, 10);
	if (res != 10) {
		sprintf(status, "TMBlockstream: %d\n", res);
	}
	return res;
}

static short select_device(short *rom)
{
	short res;
	res = TMRom(session_handle, state_buffer, rom);
	if (res != 1) {
		sprintf(status, "Can't set rom address\n");
		return -1;
	}
	res = TMStrongAccess(session_handle, state_buffer);
	if (res == 0) {
		sprintf(status, "Device not on network\n");
		return -1;
	} else if (res < 0) {
		sprintf(status, "TMStrongAccess: %d\n", res);
		return -1;
	}

	sprintf(status, "Device selected for operation\n");
	return 0;
}

/* list all devices, optionally returning the one we're looking for */
/* this is much more complicated than it should be because tdGetName can return garbled names */
static int list_devices(char *device)
{
	int i, from, to, n, id, found = -1;
	char *name, tmp[100];
	n = tdGetNumberOfDevices();
	if (!device) printf("Total number of devices: %d\n", n);
	if (!device) printf("#\tID\tName\n");
	for (i = 0; i < n; i++) {
		if (verbose) printf("i = %d\n", i);
		id = tdGetDeviceId(i);
		if (verbose) printf("id = %d\n", id);
		if (id == -1) {
			if (!device) printf("Device %d not found\n", i);
			continue;
		}
		name = tdGetName(id);
		if (verbose) printf("id = %d\n", id);
		from = 0;
		to = 0;
		while (name[from] && to < sizeof tmp) {
			if (isascii(name[from])) {
				tmp[to++] = name[from++];
			} else {
				if (verbose) printf("Non-ascii character %d in device name\n", name[from]);
				from++;
			}
		}
		if (to < sizeof tmp) tmp[to] = '\0';
		if (verbose) printf("id = %d\n", id);
		if (verbose) printf("name = '%s'\n", tmp);
		if (!device) printf("%d\t%d\t'%s'\n", i, id, tmp);
		if (verbose) printf("id = %d\n", id);
		if (device != NULL && !strncmp(device, tmp, sizeof tmp)) found = id;
		//tdReleaseString(name);
		Sleep(10);	/* not sure why */
	}
	if (verbose) printf("found = %d\n", found);
	Sleep(100);
	return found;
}

// The switch code is quite temperamental, add delays to keep it calm.
static void switch_on(int id)
{
	Sleep(100);
	tdTurnOn(id);
	Sleep(100);
}

static void switch_off(int id)
{
	Sleep(100);
	tdTurnOff(id);
	Sleep(100);
}

static int read_ds18b20(double *dtemp)
{
	short portnum;
	short porttype;
	short res;
	int i;
	short rom[8];
	//long options = 0;
	char buf[200];
	//FileEntry fentry;
	short stemp;

	if (verbose) printf("read_ds18b20(%p)\n", dtemp);
	TMReadDefaultPort(&portnum, &porttype);
	sprintf(status, "portnum = %d, porttype = %d\n", portnum, porttype);
	if (verbose) printf("%s", status);
	Get_Version(buf);
	sprintf(status, "driver: %s\n", buf);
	if (verbose) printf("%s", status);
	if (TMGetTypeVersion(porttype, buf) < 0) {
		sprintf(status, "No hardware driver found\n");
	if (verbose) printf("%s", status);
		return -1;
	}
	sprintf(status, "type version: %s\n", buf);
	if (verbose) printf("%s", status);
	session_handle = TMExtendedStartSession(portnum, porttype, NULL);
	if (session_handle > 0) {
		sprintf(status, "Session handle: %ld\n", session_handle);
	if (verbose) printf("%s", status);
	} else if (session_handle == 0) {
		sprintf(status, "Port not available\n");
	if (verbose) printf("%s", status);
		return -1;
	} else {
		sprintf(status, "Error code: %ld\n", session_handle);
	if (verbose) printf("%s", status);
		return -1;
	}

	res = TMSetup(session_handle);
	switch (res) {
	case 0:
		sprintf(status, "Setup failed\n");
	if (verbose) printf("%s", status);
		return -1;
	case 1:
		sprintf(status, "Setup ok\n");
	if (verbose) printf("%s", status);
		break;
	case 2:
		sprintf(status, "Setup ok but 1-wire network shorted\n");
	if (verbose) printf("%s", status);
		break;
	case 3:
		sprintf(status, "1-wire network does not exist\n");
	if (verbose) printf("%s", status);
		return -1;
	case 4:
		sprintf(status, "TMSetup not supported\n");
		return -1;
	default:
		sprintf(status, "setup error %d\n", res);
	if (verbose) printf("%s", status);
		return -1;
	}

	while ((res = TMNext(session_handle, state_buffer)) > 0) {
		rom[0] = 0;
		res = TMRom(session_handle, state_buffer, rom);
		for (i = 0; i < 8; i++) {
			sprintf(status, "%02X ", rom[i]);
	if (verbose) printf("%s", status);
		}
		sprintf(status, "\n");
	if (verbose) printf("%s", status);
		if (rom[0] == 0x28) {
			sprintf(status, "Found DS18B20 programmable resolution digital thermometer, done!\n");
	if (verbose) printf("%s", status);
			break;
		}
	}
	if (rom[0] != 0x28) {
		sprintf(status, "No DS18B20 found\n");
	if (verbose) printf("%s", status);
		return -1;
	}

	/* we have device address, select the device */
	select_device(rom);

	/* initiate temperature measurement */
	TMOneWireLevel(session_handle, 1, 0x01, 2);	// strong pull up after next TMTouchByte
	TMTouchByte(session_handle, 0x44);		// convert t
	Sleep(100);					// how long should we wait? 100 ms?
	TMOneWireLevel(session_handle, 1, 0x00, 0);	// normal pull up immediately

	/* select the device again */
	select_device(rom);
	res = read_scratchpad();
	sprintf(status, "Scratchpad contents: ");
	if (verbose) printf("%s", status);
	for (i = 0; i < res; i++) {
		sprintf(status, "%02x ", state_buffer[i]);
	if (verbose) printf("%s", status);
	}
	sprintf(status, "\n");
	if (verbose) printf("%s", status);
	stemp = (state_buffer[2] << 8)|state_buffer[1];
	*dtemp = stemp / 16.0;
	sprintf(status, "Temperature: %f C\n", *dtemp);
	if (verbose) printf("%s", status);

	TMEndSession(session_handle);

	return 0;
}

static void msg(char *outfile, char *color, char *message)
{
	FILE *fp;

	if (!outfile) fp = stderr;
	else fp = fopen(outfile, "w");
	if (!fp) fprintf(stderr, "Can't open %s for writing.\n", outfile);
	fprintf(fp, "%s\n", color);
	fprintf(fp, "%s\n", message);

	fprintf(fp, "temperature    : %f\n", temperature);
	fprintf(fp, "temp_target    : %f\n", temp_target);
	fprintf(fp, "temp_warn_high : %f\n", temp_warn_high);
	fprintf(fp, "temp_warn_low  : %f\n", temp_warn_low);
	fprintf(fp, "temp_crit_high : %f\n", temp_crit_high);
	fprintf(fp, "temp_crit_low  : %f\n", temp_crit_low);

	fclose(fp);
	tdClose();
	exit(EXIT_SUCCESS);
}

static void usage(void)
{
	printf("Usage: incubator.exe [-h]|[-l]|-i id[-t T][-d T][-w T][-a T][-o file][-v]\n");
	printf("\t-h\tDisplay this help text and quit.\n");
	printf("\t-l\tList available tellsticks, then quit.\n");
	printf("\t-i id\tTellstick id. No default.\n");
	printf("\t-d T\tTarget temperature. Default: %f C\n", TARGET_TEMP);
	printf("\t-d T\tMax temperature offset before turning power on or off. Default: %f\n", DIFF_TEMP);
	printf("\t-w T\tMax temperature offset before yellow status. Default: %f\n", WARN_TEMP);
	printf("\t-a T\tMax temperature offset before red status. Default: %f\n", ALARM_TEMP);
	printf("\t-o file\tName of report file. Default: stderr\n");
	printf("\t-v\tBe verbose. Repeat for even more verbosity.\n");
	exit(EXIT_SUCCESS);
}

int main(int argc, char **argv)
{
	double target_temp = TARGET_TEMP, diff_temp = DIFF_TEMP;
	double warn_temp = WARN_TEMP, alarm_temp = ALARM_TEMP;
	double temp_on, temp_off;
	char *color;
	char *outfile = NULL;
	int i;

	int res, opt, stick_id = -1;
	tdInit();
	Sleep(100);
	while ((opt = getopt(argc, argv, "hli:t:d:w:a:o:v")) != -1) {
		if (verbose) printf("Option '%c'\n", opt);
		switch(opt) {
		case 'h':
			usage();
			break;
		case 'l':
			list_devices(NULL);
			Sleep(100);
			return EXIT_SUCCESS;
		case 'i':
			stick_id = list_devices(optarg);
			Sleep(100);
			if (stick_id == -1) {
				printf("Switch with name '%s' not found.\n", optarg);
			}
			if (verbose) printf("Swich id: %d\n", stick_id);
			Sleep(100);
			break;
		case 't':
			target_temp = atof(optarg);
			break;
		case 'd':
			diff_temp = atof(optarg);
			break;
		case 'w':
			warn_temp = atof(optarg);
			break;
		case 'a':
			alarm_temp = atof(optarg);
			break;
		case 'o':
			outfile = optarg;
			break;
		case 'v':
			verbose++;
			printf("Verbosity = %d\n", verbose);
			break;
		}
	}
	if (verbose) printf("Done processing options\n");
	temp_target = target_temp;
	temp_off = target_temp+diff_temp;
	temp_on = target_temp-diff_temp;
	temp_warn_high = target_temp+warn_temp;
	temp_warn_low = target_temp-warn_temp;
	temp_crit_high = target_temp+alarm_temp;
	temp_crit_low = target_temp-alarm_temp;
	if (stick_id == -1) {
		msg(outfile, "red", "No switch");
	}
	color = "red";

	if (verbose) printf("Reading temperature\n");
	for (i = 0; i < 5; i++) {
		res = read_ds18b20(&temperature);
		if (verbose) printf("Status after read_ds18b20: %s\n", status);
		if (res == -1) {
			sprintf(status, "Error reading temperature, sleeping for 1 second.\n");
			Sleep(1000);
		} else {
			break;
		}
	}
	if (verbose) printf("Done reading temperature\n");
	if (res == -1) {
		msg(outfile, "red", "Error reading temperature.");
		exit(EXIT_FAILURE);
	}
	sprintf(status, "Temperature = %f C\n", temperature);
	if (temperature < temp_on) {
		sprintf(status, "Turn on heat because %f < %f", temperature, temp_on);
		switch_on(stick_id);
	} else if (temperature > temp_off) {
		sprintf(status, "Turn off heat because %f > %f", temperature, temp_off);
		switch_off(stick_id);
	} else {
		sprintf(status, "Don't do anything because %f < %f < %f",
			temp_on, temperature, temp_off);
	}
	if (temperature < temp_crit_low || temperature > temp_crit_high) {
		color = "red";
	} else if (temperature < temp_warn_low || temperature > temp_warn_high) {
		color = "yellow";
	} else {
		color = "green";
	}
	msg(outfile, color, status);

	return 0;
}
